@@ -1,5 +1,5 @@
 /*
- *  This file is part of gpw.
+ *  This file is part of gnu-pw-mgr.
  *
  *  Copyright (C) 2013-2018 Bruce Korb, all rights reserved.
  *  This is free software. It is licensed for use, modification and
@@ -49,8 +49,11 @@ make_pwid_mark(char const * name, size_t * len)
 
     {
         static unsigned long const mark_size =
-	    id_mark_fmt_LEN + MARK_TEXT_LEN + 10;
+            id_mark_fmt_LEN + MARK_TEXT_LEN + 10;
         char * mark = scribble_get(mark_size);
+        size_t altlen;
+        if (len == NULL)
+            len = &altlen;
         *len = sprintf(mark, id_mark_fmt, txtbuf);
         return mark;
     }
@@ -77,6 +80,7 @@ day_to_string(char const * day_str)
  * @param scan      current position in config file data
  * @param mark      the marker for "password id options"
  * @param mark_len  the length of that marker
+ *
  * @returns the character after the scan marker, or NULL
  */
 static char const *
@@ -118,12 +122,12 @@ next_pwid_opt(char const * scan, char const * mark, size_t mark_len)
         case SET_CMD_NO_PBKDF2:
         case SET_CMD_USE_PBKDF2:
             if (HAVE_OPT(REHASH))
-		continue;
+                continue;
 
             if (strncmp(scan, date_z, date_z_LEN) == 0)
-                pbkdf2_date = day_to_string(scan + date_z_LEN);
+                rehash_date = day_to_string(scan + date_z_LEN);
             else
-                pbkdf2_date = pw_undated;
+                rehash_date = pw_undated;
             break;
 
         case SET_CMD_SPECIALS:
@@ -161,79 +165,86 @@ next_pwid_opt(char const * scan, char const * mark, size_t mark_len)
 }
 
 /**
- * set the options for a particular password id.
+ * set the config file stored options for a particular password id.
  * It modifies the \a optCookie field of \a DESC(CCLASS).
  *
- * @param[in]  name   the password id name
+ * @param[in]  opt_text  the text of the configured value
+ *
+ * @returns a pointer to the text immediately after the option
  */
-static void
-set_pwid_opts(char const * name)
+static char *
+load_one_stored_opt(char const * opt_text)
 {
-    char const * cfg_text = load_config_file();
+    char * opt_buf;
+    size_t text_len;
+    char * end = strstr(opt_text, id_mark_end);
+    if (end == NULL)
+        die(GNU_PW_MGR_EXIT_BAD_CONFIG, no_id_mark_end, opt_text);
+
+    text_len = end - opt_text;
+    opt_buf = scribble_get(text_len + 1);
+    memcpy(opt_buf, opt_text, text_len);
+    opt_buf[text_len] = NUL;
+    optionLoadLine(&gnu_pw_mgrOptions, opt_buf);
+    return end + id_mark_end_LEN;
+}
+
+/**
+ * set the config file stored options for a particular password id.
+ * It modifies the \a optCookie field of \a DESC(CCLASS).
+ *
+ * @param[in]  mark     the marker for this password id
+ * @param[in]  mark_len the length of the marker
+ *
+ * @returns true if the PBKDF2 option was set via the rehash option
+ */
+static bool
+set_stored_opts(char * mark, size_t mark_len)
+{
+    bool res = false;
 
     /*
      * Find the marker that separates the seeds from the
      * password id options
      */
-    char const * scan = strstr(cfg_text, pw_id_tag);
+    char const * scan = config_file_text;
 
     if (HAVE_OPT(REHASH)) {
-	pbkdf2_date = pw_today;
-	OPT_VALUE_PBKDF2 = OPT_VALUE_REHASH;
+        rehash_date = pw_today;
+        DESC(PBKDF2).fOptState &= OPTST_PERSISTENT_MASK;
+        DESC(PBKDF2).fOptState |= OPTST_DEFINED;
+        OPT_VALUE_PBKDF2        = OPT_VALUE_REHASH;
+        res = true;
     }
 
-    if (scan != NULL) {
-        size_t mark_len;
-        char * mark = make_pwid_mark(name, &mark_len);
-
-        scan += pw_id_tag_LEN;
-
-        for (;;) {
-            char * end;
-            char * opt_text;
-            size_t text_len;
-
-            scan = next_pwid_opt(scan, mark, mark_len);
-            if (scan == NULL)
-                break;
-
-            end  = strstr(scan, id_mark_end);
-            if (end == NULL)
-                break;
-
-            text_len = end - scan;
-            opt_text = scribble_get(text_len + 1);
-            memcpy(opt_text, scan, text_len);
-            opt_text[text_len] = NUL;
-            optionLoadLine(&gnu_pw_mgrOptions, opt_text);
-            scan = end + id_mark_end_LEN;
-        }
+    while (scan = next_pwid_opt(scan, mark, mark_len),
+           scan != NULL) {
+        have_stored_opts = true;
+        scan = load_one_stored_opt(scan);
     }
 
-    sanity_check_cclass();
+    return res;
 }
 
 /**
- * Remove a replaced option.  A new instance of an option is about to
- * be inserted in the config file.  Remove any older instances of it.
+ * Sometimes the cclass option depends on the old value. Fix it up, in case.
  *
- * @param[in,out]   txt     the configuration text
- * @param[in]       mark    the password id hash in base64
- * @param[in]       m_len   the length of that hash
- * @param[in]       typ     the option type to remove
+ * @param[in]  buf     scan this buffer for the desired option
+ * @param[in]  mark    the password id hash in base64
+ * @param[in]  m_len   the length of that hash
+ * @param[in]  typ     the enumerated value of the searched for entry
+ *
+ * @returns a pointer to the start of the entry, or NULL
  */
-static void
-remove_opt(char const * txt, char const * mark, size_t m_len,
-           set_opt_enum_t typ)
+static char *
+search_for_option(char * buf, char const * mark, size_t m_len, set_opt_enum_t typ)
 {
-    char * buf = (char *)(void *)txt;
-
     for (;;) {
         char * popt = strstr(buf, mark);
         char * p    = popt;
 
         if (p == NULL)
-            return;
+            return NULL;
 
         /*
          * The marker may have more than just <pwtag id="..">, so
@@ -241,7 +252,7 @@ remove_opt(char const * txt, char const * mark, size_t m_len,
          */
         p = strchr(p, '>');
         if (p == NULL)
-            return;
+	    die(GNU_PW_MGR_EXIT_BAD_CONFIG, no_id_mark_end, mark);
 
         /*
          * Convert the next part into the option type enumeration
@@ -249,86 +260,242 @@ remove_opt(char const * txt, char const * mark, size_t m_len,
          * Spellings are "allowed" to vary, so it's not just a
          * strncmp().
          */
-        if (find_set_opt_cmd(++p) == typ) {
-            buf = popt;
-            break;
-        }
+        if (find_set_opt_cmd(++p) == typ)
+	    return popt;
+
         buf = p + id_mark_end_LEN;
     }
-    {
-        char * next = strstr(buf + m_len, pwtag_z);
-        if (next == NULL) {
-            while (buf[-1] == NL)  buf--;
-            *(buf++) = NL;
-            *buf = NUL;
-        } else {
-            size_t ln = strlen(next) + 1;
-            memmove(buf, next, ln);
-        }
-    }
 }
 
+/**
+ * Before removing a cclass option from the configuration text,
+ * make sure the newly defined option doesn't start with a '+' or '-'.
+ * If it does, then we are modifying a previously existing value.
+ *
+ * @param[in]  mark    the password id hash in base64
+ * @param[in]  m_len   the length of that hash
+ * @param[in]  typ     the enumerated value of the searched for entry
+ */
+static void
+adjust_cclass_val(char const * mark, size_t m_len)
+{
+    intptr_t     new_cc = (intptr_t)OPT_VALUE_CCLASS;
+    intptr_t     old_cc;
+    char *       buf =
+        search_for_option(config_file_text, mark, m_len, SET_CMD_CCLASS);
+
+    if (buf == NULL) {
+        old_cc = (intptr_t)(
+            HAVE_OPT(DEFAULT_CCLASS)
+            ? DESC(DEFAULT_CCLASS).optCookie
+            : CclassCookieBits );
+    } else {
+        char * scan = strchr(buf, '>');
+        if (scan == NULL)
+            die(GNU_PW_MGR_EXIT_BAD_CONFIG, no_id_mark_end, buf);
+
+        (void) load_one_stored_opt(scan+1);
+
+        old_cc = OPT_VALUE_CCLASS;
+    }
+
+    {
+        intptr_t new = (tweak_prev_cclass < 0)
+            ? (old_cc & ~new_cc) : (old_cc | new_cc);
+
+        DESC(CCLASS).fOptState &= OPTST_PERSISTENT_MASK;
+        DESC(CCLASS).fOptState |= OPTST_DEFINED;
+        DESC(CCLASS).optCookie  = (void*)new;
+    }
+}
+
+/**
+ * Before removing a cclass option from the configuration text,
+ * make sure the newly defined option doesn't start with a '+' or '-'.
+ * If it does, then we are modifying a previously existing value.
+ *
+ * @param[in]  mark    the password id hash in base64
+ * @param[in]  m_len   the length of that hash
+ */
+static void
+adjust_pbkdf2_val(char const * mark, size_t m_len)
+{
+    uint64_t     old_pbkdf2 = (intptr_t)PBKDF2_DFT_ARG;
+    uint64_t     new_pbkdf2 = OPT_VALUE_REHASH;
+    char *       buf =
+        search_for_option(config_file_text, mark, m_len, SET_CMD_USE_PBKDF2);
+
+    /*
+     * If there is an entry, then pull it out and stash the "old" value.
+     */
+    if (buf != NULL) {
+        char * scan = strchr(buf, '>');
+        if (scan == NULL)
+            die(GNU_PW_MGR_EXIT_BAD_CONFIG, no_id_mark_end, buf);
+
+        (void) load_one_stored_opt(scan+1);
+
+        old_pbkdf2 = OPT_VALUE_PBKDF2;
+    }
+
+    /*
+     * Now adjust the value and see if it has changed.
+     */
+    {
+        uint64_t new = (tweak_prev_rehash < 0)
+            ? (old_pbkdf2 - new_pbkdf2) : (old_pbkdf2 + new_pbkdf2);
+
+        if (new == 0) {
+            new = (uint64_t)(intptr_t)PBKDF2_DFT_ARG;
+            warning_msg(rehash_set_fmt, new);
+
+        } else if (new > (uint64_t)100000) {
+            new = 1;
+            warning_msg(rehash_set_fmt, new);
+        }
+
+        DESC(PBKDF2).fOptState &= OPTST_PERSISTENT_MASK;
+        DESC(PBKDF2).fOptState |= OPTST_DEFINED;
+        OPT_VALUE_PBKDF2        = new;
+    }
+}
+
+/**
+ * Sometimes the cclass option depends on the old value. Fix it up, in case.
+ *
+ * @param[in]  mark    the password id hash in base64
+ * @param[in]  m_len   the length of that hash
+ *
+ * @returns true if the option was actually removed
+ */
 static bool
-removed_old_opts(char const * cfg_text, char const * name, char ** mark_p)
+remove_opt(char const * mark, size_t m_len, set_opt_enum_t typ)
+{
+    char * buf = config_file_text;
+
+    while (buf = search_for_option(buf, mark, m_len, typ),
+	   buf != NULL) {
+
+	char * next = strstr(buf + m_len, id_mark_end);
+	if (next == NULL)
+	    die(GNU_PW_MGR_EXIT_BAD_CONFIG, no_id_mark_end, buf);
+
+	next += id_mark_end_LEN;
+	while (*next == NL) next++;
+
+	if (*next == NUL) {
+	    *buf = NUL;
+	    break;
+	}
+
+	memmove(buf, next, strlen(next) + 1);
+    }
+
+    return true;
+}
+
+/**
+ * check for updated options
+ *
+ * @param mark     the string that marks all options for the current password id
+ * @param mark_len the length of that string
+ *
+ * @returns true if any stored options need updating.
+ */
+static bool
+remove_defined_opts(char * mark, size_t mark_len)
 {
     bool res = false;
-    size_t mark_len;
 
-    char * mark = *mark_p = make_pwid_mark(name, &mark_len);
+    if (STATE_OPT(LOGIN_ID) == OPTST_DEFINED)
+        res |= remove_opt(mark, mark_len, SET_CMD_LOGIN_ID);
 
-    if (STATE_OPT(LOGIN_ID) == OPTST_DEFINED) {
-        res = true;
-        remove_opt(cfg_text, mark, mark_len, SET_CMD_LOGIN_ID);
-    }
-
-    if (STATE_OPT(LENGTH) == OPTST_DEFINED) {
-        res = true;
-        remove_opt(cfg_text, mark, mark_len, SET_CMD_LENGTH);
-    }
+    if (STATE_OPT(LENGTH) == OPTST_DEFINED)
+        res |= remove_opt(mark, mark_len, SET_CMD_LENGTH);
 
     if (STATE_OPT(CCLASS) == OPTST_DEFINED) {
-        res = true;
-        remove_opt(cfg_text, mark, mark_len, SET_CMD_CCLASS);
+        if (tweak_prev_cclass != 0) {
+            adjust_cclass_val(mark, mark_len);
+            res = true;
+        }
+        res |= remove_opt(mark, mark_len, SET_CMD_CCLASS);
     }
+
     if (HAVE_OPT(REHASH)) {
-        res = true;
-        remove_opt(cfg_text, mark, mark_len, SET_CMD_NO_PBKDF2);
-        remove_opt(cfg_text, mark, mark_len, SET_CMD_USE_PBKDF2);
+        if (tweak_prev_rehash != 0) {
+            adjust_pbkdf2_val(mark, mark_len);
+            res = true;
+        }
+        res |= remove_opt(mark, mark_len, SET_CMD_NO_PBKDF2);
+        res |= remove_opt(mark, mark_len, SET_CMD_USE_PBKDF2);
     }
 
-    if (STATE_OPT(SPECIALS) == OPTST_DEFINED) {
-        res = true;
-        remove_opt(cfg_text, mark, mark_len, SET_CMD_SPECIALS);
-    }
+    if (STATE_OPT(SPECIALS) == OPTST_DEFINED)
+        res |= remove_opt(mark, mark_len, SET_CMD_SPECIALS);
 
-    if (STATE_OPT(SHARED) == OPTST_DEFINED) {
-        res = true;
-        remove_opt(cfg_text, mark, mark_len, SET_CMD_SHARED);
-    }
+    if (STATE_OPT(SHARED) == OPTST_DEFINED)
+        res |= remove_opt(mark, mark_len, SET_CMD_SHARED);
 
     return res;
 }
 
+/**
+ * set the options for a particular password id.
+ * It modifies the \a optCookie field of \a DESC(CCLASS).
+ *
+ * @param[in]  pw_id   the password id
+ */
+static void
+set_pwid_opts(char const * pw_id)
+{
+    size_t mark_len;
+    char const * scan = config_file_text;
+    char *       mark = make_pwid_mark(pw_id, &mark_len);
+
+    /*
+     * Get rid of any stored options that appear on the command line
+     */
+    update_stored_opts |= remove_defined_opts(mark, mark_len);
+
+    /*
+     * now set all the options specified in the config file.
+     * None will conflict with command line options since we just
+     * removed all the conflicts with the ones stored.
+     */
+    update_stored_opts |= set_stored_opts(mark, mark_len);
+
+    /*
+     * If we have a default character class and we did not find
+     * any stored options for this password id, then set the
+     * caracter classes to the default and note that we must
+     * rewrite the cclass option. (Setting state to "DEFINED"
+     * tells the wrap up code that the value must be saved.)
+     */
+    if ((! HAVE_OPT(CCLASS)) && HAVE_OPT(DEFAULT_CCLASS)) {
+        SET_OPT_CCLASS((uintptr_t) (void*) OPT_ARG(DEFAULT_CCLASS));
+        DESC(CCLASS).fOptState &= OPTST_PERSISTENT_MASK;
+        DESC(CCLASS).fOptState |= OPTST_DEFINED;
+        update_stored_opts      = true;
+    }
+    if (HAVE_OPT(CCLASS))
+        sanity_check_cclass();
+}
+
 /**
  * Update password specific options.  The password-options must be
- * checked for being "defined" (set on the command line).  If they
- * are, remove them from the config data and append the new value.
+ * checked for being "defined" (set on the command line).
  *
  * @param  name  password id
  */
 static void
 update_pwid_opts(char const * name)
 {
-    char const * cfg_text = load_config_file();
-    char const * scan     = strstr(cfg_text, pw_id_tag);
-    char * mark = NULL;
-
-    if (scan == NULL) {
-        size_t len = strlen(cfg_text);
+    if (strstr(config_file_text, pw_id_tag) == NULL) {
+        size_t len = strlen(config_file_text);
         char * emk = scribble_get(len + pw_id_tag_LEN + 3);
 
-        memcpy(emk, cfg_text, len);
-        cfg_text = emk;
+        memcpy(emk, config_file_text, len);
+        config_file_text = emk;
         emk += len;
 
         memcpy(emk, pw_id_tag, pw_id_tag_LEN);
@@ -338,20 +505,18 @@ update_pwid_opts(char const * name)
         *emk     = NUL;
     }
 
-    if (! removed_old_opts(cfg_text, name, &mark))
-        return;
-
     /*
-     * We have new info to stash.  Any old values in "cfg_text"
+     * We had at least one command line option.
      */
     {
+        char * mark = make_pwid_mark(name, NULL);
         char const * fnm = access_config_file();
         FILE * fp = fopen(fnm, "w");
 
         if (fp == NULL)
             fserr(GNU_PW_MGR_EXIT_NO_CONFIG, fopen_z, fnm);
 
-        fputs(cfg_text, fp);
+        fputs(config_file_text, fp);
         if (STATE_OPT(LOGIN_ID) == OPTST_DEFINED)
             fprintf(fp, pwid_login_id_fmt, mark, OPT_ARG(LOGIN_ID));
 
@@ -359,7 +524,12 @@ update_pwid_opts(char const * name)
             fprintf(fp, pwid_length_fmt, mark, (unsigned int)OPT_VALUE_LENGTH);
 
         if (STATE_OPT(CCLASS) == OPTST_DEFINED) {
-            tOptDesc *   od   = gnu_pw_mgrOptions.pOptDesc + INDEX_OPT_CCLASS;
+            /*
+             * The CCLASS is specified as a series of bits. Call the option
+             * handling function with a magic address for the option structure
+             * and it will replace the binary value with an allocated string.
+             */
+            tOptDesc *   od   = &DESC(CCLASS);
             char const * save = od->optArg.argString;
             doOptCclass(OPTPROC_RETURN_VALNAME, od);
             fprintf(fp, pwid_cclass_fmt, mark, od->optArg.argString);
@@ -367,12 +537,21 @@ update_pwid_opts(char const * name)
             od->optArg.argString = save;
         }
 
-	if (HAVE_OPT(REHASH)) {
+        /*
+         * We are here because a new persistent option was specified.
+         *
+         * If either --rehash was specified *OR*
+         *    some other persistent option was specified,
+         * then we'll stash the rehash value and timestamp the password id
+         *
+         * NOTE CAREFULLY: if there is a previous rehash value, then
+         * "have_stored_opts" will be true and we won't update the date.
+         */
+        if (HAVE_OPT(REHASH) || (! have_stored_opts)) {
             unsigned int day = (unsigned int)
                 (time(NULL) / SECONDS_IN_DAY);
-
-            fprintf(fp, pwid_pbkdf2_fmt, mark, day,
-		    (uint32_t)OPT_VALUE_REHASH);
+	    uint32_t val = HAVE_OPT(REHASH) ? OPT_VALUE_REHASH : OPT_VALUE_PBKDF2;
+            fprintf(fp, pwid_pbkdf2_fmt, mark, day, val);
         }
 
         if (STATE_OPT(SPECIALS) == OPTST_DEFINED)
@@ -392,33 +571,36 @@ update_pwid_opts(char const * name)
 static void
 remove_pwid(char const * name)
 {
+    fwrite(rm_entry, rm_entry_LEN, 1, stdout);
     print_pwid_status(name);
     {
-        size_t       mark_len;
-        char const * cfg_text = load_config_file();
-        char *       scan     = strstr(cfg_text, pw_id_tag);
-        char *       mark     = make_pwid_mark(name, &mark_len);
         bool         found    = false;
+        size_t       mark_len;
+        char *       mark     = make_pwid_mark(name, &mark_len);
+        char *       scan     = config_file_text;
 
-        if (scan == NULL)
-            return;
-        scan += pw_id_tag_LEN;
-
-        while (scan = strstr(scan + 1, mark),
+        while (scan = strstr(scan, mark),
                scan != NULL) {
             char * sol = scan;
             found = true;
 
-        find_line_end:
+        find_next_tag_end:
 
+            /*
+             * If we can't find a tag end marker, truncate the file here.
+             */
             scan = strstr(scan + mark_len, pwtag_z);
             if (scan == NULL) {
                 *sol = NUL;
                 break;
             }
 
-            if (strncmp(scan, mark, mark_len) == 0)
-                goto find_line_end;
+            if (   (scan[pwtag_z_LEN] == NL)
+                && (strncmp(scan + pwtag_z_LEN + 1, mark, mark_len) == 0))
+                {
+                    scan += pwtag_z_LEN + 1;
+                    goto find_next_tag_end;
+                }
 
             memmove(sol, scan, strlen(scan) + 1);
             scan = sol;
@@ -431,7 +613,7 @@ remove_pwid(char const * name)
             if (fp == NULL)
                 fserr(GNU_PW_MGR_EXIT_NO_CONFIG, fopen_z, fnm);
 
-            fputs(cfg_text, fp);
+            fputs(config_file_text, fp);
             fclose(fp);
         }
     }
@@ -522,12 +704,11 @@ create_cfg_file(char * opt, char * opt1)
 static void
 fix_options(int * ac, char *** av)
 {
-    static char const conf[] = "--config";
     int     argc  = *ac;
     char ** argv  = *av;
     while (--argc > 0) {
         char * a = *++argv;
-        int    c = strncmp(a, conf, sizeof(conf) - 1);
+        int    c = strncmp(a, dash_config_z, dash_config_z_LEN);
         if (c == 0) {
             create_cfg_file(a, argv[1]);
             return;
@@ -535,3 +716,11 @@ fix_options(int * ac, char *** av)
     }
     insert_load_opts(ac, av);
 }
+
+/*
+ * Local Variables:
+ * mode: C
+ * c-file-style: "stroustrup"
+ * indent-tabs-mode: nil
+ * End:
+ * end of pw-opts.c */
