@@ -1,7 +1,7 @@
 /*
- *  This file is part of gpw.
+ *  This file is part of gnu-pw-mgr.
  *
- *  Copyright (C) 2013-2016 Bruce Korb, all rights reserved.
+ *  Copyright (C) 2013-2018 Bruce Korb, all rights reserved.
  *  This is free software. It is licensed for use, modification and
  *  redistribution under the terms of the GNU General Public License,
  *  version 3 or later <http://gnu.org/licenses/gpl.html>
@@ -124,7 +124,7 @@ ver_str_to_number(void)
             die(GNU_PW_MGR_EXIT_CODING_ERROR, bad_vers);
         res += v << shift;
         switch (*pn) {
-        case '-': case '\0':
+        case '-': case NUL:
             goto return_res;
 
         case '.':
@@ -143,25 +143,35 @@ ver_str_to_number(void)
     return res;
 }
 
-static inline void
-print_new_seed(FILE * fp, char const * cfg_text)
+/**
+ * add or replace the default character class for new passwords
+ */
+static void
+set_default_cclass(void)
 {
-    uint32_t     seed_ver = ver_str_to_number();
-    char const * seed_txt = get_seed_text();
+    char const * cfg_file;
+    load_config_file();
+    cfg_file = access_config_file();
+    FILE * fp = fopen(cfg_file, "w");
+    if (fp == NULL)
+	fserr(GNU_PW_MGR_EXIT_NO_CONFIG, fopen_z, cfg_file);
 
-    if (HAVE_OPT(SHARED)) {
-        fprintf(fp, cfg_fmt, OPT_ARG(TAG), seed_ver, sec_mark, seed_txt);
-        fputs(cfg_text, fp);
+    fprintf(fp, default_cclass_fmt, OPT_ARG(DEFAULT_CCLASS));
+    do {
+	char const * old_cc = strstr(config_file_text, default_cclass);
+	if (old_cc == NULL)
+	    break;
 
-    } else {
-        char const * rest_of_text = strstr(cfg_text, pw_id_tag);
-        fprintf(fp, cfg_fmt, OPT_ARG(TAG), seed_ver, "", seed_txt);
+	/* skip over the previous default_cclass */
+	fwrite(config_file_text, config_file_text - old_cc, 1, fp);
+	old_cc = strstr(old_cc + default_cclass_LEN, default_cclass+1);
+	if (old_cc == NULL)
+	    fserr(GNU_PW_MGR_EXIT_NO_CONFIG, fopen_z, cfg_file);
+	fputs(old_cc + default_cclass_LEN - 1, fp);
+    } while (0);
 
-        if (rest_of_text != NULL)
-            fwrite(cfg_text, rest_of_text - cfg_text, 1, fp);
-        else
-            fputs(cfg_text, fp);
-    }
+    fputs(config_file_text, fp);
+    fclose(fp);
 }
 
 /**
@@ -171,25 +181,34 @@ print_new_seed(FILE * fp, char const * cfg_text)
 static void
 add_seed(void)
 {
-    char const * cfg_text = load_config_file();
+    char const * cfg_file;
     FILE * fp;
 
+    load_config_file();
+    cfg_file = access_config_file();
+    fp = fopen(cfg_file, "w");
+    if (fp == NULL)
+	fserr(GNU_PW_MGR_EXIT_NO_CONFIG, fopen_z, cfg_file);
+
+    /*
+     * The new tag must be unique. Delete the old one first.
+     */
     {
-        char * tag = scribble_get(sizeof (tag_fmt) + strlen(OPT_ARG(TAG)));
+        char * tag = scribble_get(tag_fmt_LEN + strlen(OPT_ARG(TAG)));
         sprintf(tag, tag_fmt, OPT_ARG(TAG));
-        if (strstr(cfg_text, tag) != NULL)
+        if (strstr(config_file_text, tag) != NULL)
             die(GNU_PW_MGR_EXIT_BAD_SEED, dup_tag, OPT_ARG(TAG));
     }
 
     {
-        char const * cfg_file = access_config_file();
-        fp = fopen(cfg_file, "w");
-        if (fp == NULL)
-            fserr(GNU_PW_MGR_EXIT_NO_CONFIG, fopen_z, cfg_file);
+	uint32_t     seed_ver = ver_str_to_number();
+	char const * seed_txt = get_seed_text();
+	char const * marker   = HAVE_OPT(SHARED) ? sec_mark : "";
+
+	fprintf(fp, cfg_fmt, OPT_ARG(TAG), seed_ver, marker, seed_txt);
     }
 
-    print_new_seed(fp, cfg_text);
-
+    fputs(config_file_text, fp);
     fclose(fp);
 }
 
@@ -199,13 +218,14 @@ add_seed(void)
 static void
 rm_seed(void)
 {
-    char const * cfg_data = load_config_file();
-    char const * prune    = cfg_data;
-
+    char const * prune;
     char * tag = scribble_get(tag_fmt_LEN + strlen(OPT_ARG(TAG)) + 1);
 
+    load_config_file();
+    prune = config_file_text;
+
     sprintf(tag, tag_fmt, OPT_ARG(TAG));
-    tag = strstr(cfg_data, tag);
+    tag = strstr(config_file_text, tag);
     if (tag == NULL)
         die(GNU_PW_MGR_EXIT_BAD_SEED, tag_gone_fmt, OPT_ARG(TAG));
 
@@ -224,8 +244,8 @@ rm_seed(void)
         FILE * fp = fopen(cfg_file, "w");
         if (fp == NULL)
             fserr(GNU_PW_MGR_EXIT_NO_CONFIG, open_z, cfg_file);
-        if (prune > cfg_data)
-            fwrite(cfg_data, prune - cfg_data, 1, fp);
+        if (prune > config_file_text)
+            fwrite(config_file_text, prune - config_file_text, 1, fp);
 
         /*
          * If there is another <seed>, print from there.
